import fs from "fs"
import path from "path"
import { fileURLToPath } from 'url';

import dayjs from "../utils/dayjs.js"

import { ECU, IOS, System, Serial, Utils } from "syrus4-nodejs"

import storage from "../services/storage.js"
import { Queue } from "../services/Queue.js"

const __filename = fileURLToPath(import.meta.url);
const __dirname = path.dirname(__filename);

const ECU_LIST = ECU.getECUList();

// lengths of the TAIP variables to filled with commas when no middle data received
const LENGTHS = {
	ZA: 9,
	ZE: 9,
	ZS: 11,
	ZH: 5,
	ZX: 7,
	ZY: 7,
};

// Helper function to construct IX= extended tag
function getIO(data) {
	let byte1 = "1";
	let byte2 = `${data["$io.ign"]}` == "true" ? 2 : 0;
	byte2 += `${data["$io.pwr"]}` == "true" ? 1 : 0;
	let byte3 = `${data["$io.in1"]}` == "true" ? 1 : 0;
	byte3 += `${data["$io.in2"]}` == "true" ? 2 : 0;
	byte3 += `${data["$io.in3"]}` == "true" ? 4 : 0;
	let byte4 = `${data["$io.out1"]}` == "true" ? 1 : 0;
	byte4 += `${data["$io.out2"]}` == "true" ? 2 : 0;

	// Extendeds
	let byte5 = `${data["$io.so1"]}` == "true" || `${data["$io.so2"]}` == "true" ? 1 : 0;
	let byte6 = `${data["$io.in4"]}` == "true" ? 1 : 0;
	byte6 += `${data["$io.in5"]}` == "true" ? 2 : 0;
	byte6 += `${data["$io.in6"]}` == "true" ? 4 : 0;
	byte6 += `${data["$io.in7"]}` == "true" ? 8 : 0;
	let byte7 = `${data["$io.out3"]}` == "true" ? 1 : 0;
	byte7 += `${data["$io.out4"]}` == "true" ? 2 : 0;
	let byte8 = `${data["$io.so3"]}` == "true" || `${data["$io.so4"]}` == "true" ? 1 : 0;
	return `;IX=1${byte1}2${byte2}3${byte3}4${byte4}5${byte5}6${byte6}7${byte7}8${byte8}`;
}

// Helper function to construct AW= extended tag
function getAnalogs(data) {
	let AD = "";
	let an1 = parseInt(data["$io.an1"]) || 0
	let an2 = parseInt(data["$io.an2"]) || 0
	let d12 = parseInt(data["$io.dan"]) || 0
	if (an1) AD = `;AD=${an1}`
	return `${AD};AW=${an1},${an2},,,,,,,${d12},,,,,,,`;
}

// TPMS
function TPMS_build_position(axle, tire){
	let axle_b = Number(axle).toString(2)
	axle_b = axle_b.padStart(4, "0")
	let tire_b = Number(tire).toString(2)
	tire_b = tire_b.padStart(4, "0")
	let position = `${parseInt(axle_b + tire_b, 2)}`
	return position.padStart(3, "0")
}
function TPMS_provisioning(value) {
	let result = []
	let tokens = value.split(';')

	let axles = new Set()
	for (const token of tokens) {
		let [axle, tire, value] = token.split(",")
		axles.add(axle)
		let position = TPMS_build_position(axle, tire)
		result.push(`${position}${Number(value).toFixed(0)}`)
	}
	let p_axles = axles.size.toString().padStart(2, '0')
	let p_tires = tokens.length.toString().padStart(2, '0')
	result.unshift(`00${p_axles}${p_tires}`)
	return result.join(",")
}
function TPMS_temperature(value) {
	let result = []
	let tokens = value.split(';')
	for (token of tokens) {
		let [axle, tire, value] = token.split(",")
		let position = TPMS_build_position(axle, tire)
		value = Number(value)
		result.push(`${position}${value.toFixed(0)}`)
	}
	return result.join(",")
}
function TPMS_pressure(value){
	let result = []
	let tokens = value.split(';')
	for (const token of tokens){
		let [axle, tire, value] = token.split(",")
		let position = TPMS_build_position(axle, tire)
		value = Number(value) / 6.89476
		result.push(`${position}${value.toFixed(0)}`)
	}
	return result.join(",")
}
function TPMS_default(value) {
	let result = []
	let tokens = value.split(';')
	for (const token of tokens) {
		let [axle, tire, value] = token.split(",")
		let position = TPMS_build_position(axle, tire)
		value = Number(value)
		result.push(`${position}${value.toFixed(0)}`)
	}
	return result.join(",")
}

// Helper function to contruct all the extended tag associated with ECU monitor
function ECUToTAIP(data) {
	let txt = "";
	const combined = {}
	const complex = {}
	const ecuFlag = data["$io.ign"] ? "T" : "F"
	for (const key in data) {
		if (!key.startsWith("$ecu.")) continue

		let value = data[key];
		const element = ECU_LIST[key.replace("$ecu.", "")]

		if (!(element && element.$taip_tag)) continue
		if (!element.$type) element.$type = 'number'
		if (!element.$taip_multiplier) element.$taip_multiplier = 1
		if (element.$type == 'number'){
			let typed = Number(value) * element.$taip_multiplier
			if(!element.$taip_float) typed = parseInt(typed)
			if(!isNaN(typed)) value = typed
		}

		let { $taip_tag } = element
		if (!element.$tokenizer && $taip_tag.includes(".")) {
			const [tag, pos] = $taip_tag.split(".");
			if (!combined[tag]) {
				combined[tag] = []
				if (LENGTHS[tag] > combined[tag].length)
					combined[tag].length = LENGTHS[tag]
					combined[tag].fill(undefined)
			}
			if (!LENGTHS[tag] && pos > combined[tag].length) combined[tag].length = pos
			combined[tag][pos - 1] = value;
			continue
		}
		if (element.$name.startsWith('tires_')){
			if (!value.includes(";")) continue
			// console.log("TPMS_DEBUG", {key, value})
			// let pre = value
			if (element.$name == 'tires_provisioning') value = TPMS_provisioning(value)
			else if (element.$name == 'tires_pressure') value = TPMS_pressure(value)
			else if (element.$name == 'tires_temperature') value = TPMS_temperature(value)
			else value = TPMS_default(value)
			if ($taip_tag.includes('.')){
				// console.log({$taip_tag, pre, value})
				const [tag, pos] = $taip_tag.split(".");
				if (!complex[tag]) complex[tag] = {}
				value.split(",").map((tire) => {
					const [tpos, tvalue] = [tire.slice(0,3), tire.slice(3)]
					if (!complex[tag][tpos]){
						complex[tag][tpos] = []
						if (LENGTHS[tag] > complex[tag][tpos].length)
							complex[tag][tpos].length = LENGTHS[tag]
							complex[tag][tpos].fill(0)
					}
					if (!LENGTHS[tag] && pos > complex[tag][tpos].length) complex[tag][tpos].length = pos
					complex[tag][tpos][pos - 1] = tvalue;
				})
				continue
			}
		}
		txt += `;${$taip_tag}=${ecuFlag}${value}`;
	}
	for (const tag in combined) {
		let arr = combined[tag];
		txt += `;${tag}=${arr
			.map(v => {
				if (v != undefined) return `${ecuFlag}${v}`;
				else return "";
			})
			.join(",")}`;
		continue
	}
	for (const tag in complex){
		let item = complex[tag]
		let sval = []
		for (const subtag in item){
			let value = item[subtag]
			if(tag == 'ZY') {
				value[2] = "3"
				value[6] = "1"
			}
			sval.push(`${subtag}${value.join("")}`)
		}
		txt += `;${tag}=${ecuFlag}${sval.join(",")}`
	}
	let error_pgn = data['$ecu.feca_3-6']
	if (error_pgn){
		let { spn, fmi, cm, oc } = {
			spn: data['$ecu.error_codes.spn'] || '0',
			fmi: data['$ecu.error_codes.fmi'] || '0',
			cm: data['$ecu.error_codes.cm'] || '0',
			oc: data['$ecu.error_codes.oc'] || '0',
		}
		let partial = `;ZC=${ecuFlag}1$${spn},${fmi},${cm},${oc}`
		if (!partial.endsWith('$0,0,0,0')) txt += partial
	}
	let mil_indicator = data['$ecu.feca_1.7']
	if (mil_indicator){
		txt += `;YT=${ecuFlag}${mil_indicator},1,0`;
	}
	// console.log({ combined, complex, txt })
	return txt;
}

function getCounters(fields, data) {
	let txt = "";
	if (fields["VO"] && data[fields["VO"]]) {
		txt += `;VO=${data[fields["VO"]]}`;
	}
	if (fields["CE"] && data[fields["CE"]]) {
		txt += `;CE=${data[fields["CE"]]}`;
	}
	if (fields["CL"] && data[fields["CL"]]) {
		txt += `;CL=${data[fields["CL"]]}`;
	}
	if (fields["CS"] && data[fields["CS"]]) {
		txt += `;CS=${data[fields["CS"]]}`;
	}
	if (fields["CR"] && data[fields["CR"]]) {
		txt += `;CR=${data[fields["CR"]]}`;
	}
	for (const key in fields) {
		if (key.startsWith("CV") && data[fields[key]]) {
			txt += `;${key}=${data[fields[key]]}`;
		}
	}
	return txt;
}

// TODO: no uppercasing here..
async function getSC(fields, data) {
	try {
		if (fields['$modem']) {
			let $modem = {
				RAT: data['$modem.rat'],
				MCC_MNC: data['$modem.mcc_mnc'],
				LAC: data['$modem.lac'],
				CID: data['$modem.cid'],
				RSSI: data['$modem.rssi'],
			}
			let { RAT, MCC_MNC, LAC, CID, RSSI } = $modem;
			let [MCC, MNC] = [MCC_MNC.slice(0, 3), MCC_MNC.slice(3)];
			CID = parseInt(CID, 16);
			if (isNaN(CID)) CID = '';
			let result = `;SC=${RAT},${MCC},${MNC},${LAC},${CID},${RSSI}`
			result = result.replace(/(undefined|null)/g, '')
			if (result == ';SC=,,,,') return ""
			return result
		}
		return ""
	} catch (e) {
		console.error(e)
		return ""
	}
}

const _satcom_cache = {}
async function getSatcom(fields, data){
	if (_satcom_cache.value) return _satcom_cache.value
	try {
		if (!fields['$satcom']) {
			return ""
		}
		let state = await Serial.getSerialModemState()
		let is_satcom = state.type == "satcom"
		_satcom_cache.value = is_satcom ? `;II=${state.id}`: ""
		// clear cache periodically (15 mins)
		setTimeout(() => { _satcom_cache.value = null }, 1000*60*15)
		return _satcom_cache.value
	} catch (e) {
		console.error(e)
		return ""
	}
}

// IButton
function getIB_IS(fields, data) {
	const empty = "0".repeat(16)
	try {
		let txt = "";
		let {IB, IS} = fields
		if (IB){
			let IB_value = data[IB] || empty
			txt += `;IB=${IB_value}`
		}
		if (IS) {
			let IS_value = data[IS] || empty
			let tokens = IS.split('.')
			tokens[tokens.length - 1] = 'connected'
			let is_connected = data[tokens.join(".")] && 1 || 0
			txt += `;IS=${is_connected}${IS_value}`
		}
		return txt;
	} catch(e){
		console.error(e)
		return ""
	}
}

// Tempearture
function getEA(fields, data){
	try {
		let txt = "";
		let { EA_A, EA_B, EA_C } = fields
		// if none are present, skip tag
		if (!(EA_A || EA_B || EA_C)){
			return txt
		}
		let tokens = []
		if (EA_A && data[EA_A]) {
			tokens.push(`A${data[EA_A]}`)
		}else{
			tokens.push("A*")
		}
		if (EA_B && data[EA_B]) {
			tokens.push(`B${data[EA_B]}`)
		} else {
			tokens.push("B*")
		}
		if (EA_C && data[EA_C]) {
			tokens.push(`C${data[EA_C]}`)
		} else {
			tokens.push("C*")
		}
		txt = `;EA=${tokens.join("")}`
		return txt;
	} catch (e) {
		console.error(e)
		return ""
	}
}

function pad(number, digits, is_decimal = false) {
	if (String(number).length > digits) {
		number = String(number).substring(0, digits);
	}
	if (is_decimal) {
		return String(number) + Array(Math.max(digits - String(number).length + 1, 0)).join("0");
	}
	return Array(Math.max(digits - String(number).length + 1, 0)).join("0") + number;
}

/**
 * Set Latitude data in  Taip/Syrus format data
 *
 * @param {number} lat
 * @returns Latitude string parser
 *
 * @memberOf Parser
 */
function getLat(lat = 0.0) {
	let integer = Math.floor(Math.abs(lat));
	let symbol = lat > 0 ? "+" : "-";
	let decimal = (lat + "").split(".")[1];
	if (!decimal) decimal = "0";
	return `${symbol}${pad(integer, 2)}${pad(decimal, 5, true)}`;
}

/**
 * Set longitude data in  Taip/Syrus format data
 *
 * @param {number} long
 * @returns longitude string parser
 *
 * @memberOf Parser
 */
function getLong(long = 0.0) {
	let integer = Math.floor(Math.abs(long));
	let symbol = long > 0 ? "+" : "-";
	let decimal = (long + "").split(".")[1];
	if (!decimal) decimal = "0";
	return `${symbol}${pad(integer, 3)}${pad(decimal, 5, true)}`;
}

function __noop(){}

/**
 * TAIP class manage all the communication in TAIP protocol and delivers order to syrusjs from TAIP commands
 */
export class TAIP {
	// events
	constructor({ name, client, ack, onCommand, autoDequeue = false }) {
		let that = this
		this.retry = null;
		this.retries = 0;
		this.name = name;
		this.queue = new Queue(`peg_queue_${name}`);
		this.queue.ready().then(list => {});
		this.connected = false;
		this.imei;
		this.IMSI;
		this.ICCID;
		this.carrier;
		this.trips_events = false;
		this.onCommand = onCommand || _noop;
		this.ack = ack || 'imei'
		this.ack_counter = 0
		this.client = client
		this.taip_buffer = ''
		this.taip_buffer_timeout = null
		this.__ready = null
		this.ready = new Promise((resolve)=>{ that.__ready = resolve })
		this.init();
	}

	// Preparing data for and connections callbacks
	async init() {
		let modem, info, version
		try {
			info = await System.info();
		} catch (error){
			console.error(`TCP:System.info()`, error);
			info = {}
		}
		try{
			modem = await System.modem();
		} catch (error) {
			console.error("TCP:System.modem()", error);
			modem = {};
		}
		// Getting the active sim
		this.imei = modem.IMEI;
		if (modem.SIM_STATE == "READY") {
			this.IMSI = modem.SIM_IMSI || modem.eSIM_IMSI || "";
			this.ICCID = modem.SIM_ID || modem.eSIM_ID || "";
		} else {
			this.IMSI = modem.eSIM_IMSI || modem.SIM_IMSI || "";
			this.ICCID = modem.eSIM_ID || modem.SIM_ID || "";
		}
		// Getting version and carrier
		version = info.apexVersion ? info.apexVersion.replace("apex-", "") : "00.00.00";
		version = version.split("-")[0] + ".00";
		this.carrier = modem.OPERATOR || "unknown";

		// ack
		this.ack_counter = await storage.get(`${this.name}_ack`);
		if (!this.ack_counter) this.ack_counter = 0

		// FTP
		this.photo_state = await storage.hgetall(`${this.name}_photo_state`)
		this.ftp = await storage.hgetall(`${this.name}_ftp_config`)
		console.log(`tcp(${this.name}):ftp`, this.ftp)

		// Establish the callbacks for the channel (TCP, HTTP, file, etc)
		this.client.onConnect = () => {
			this.connected = true;
			// some clients (i.e satcom, don't need a welcome message)
			if(this.client.disable_welcome_message){
				this._sendBuffer()
				return
			}
			// Sending salute after connection
			let package_json = JSON.parse(fs.readFileSync(path.resolve(__dirname, "../package.json")))
			let additional = ''
			additional += `model=Syrus 4 LTE,syrusjs_version=${package_json.version}`
			additional += `,os_version=${info.apexVersion}`
			additional += `,kernel_version=${info.kernelVersion},platform=APEX OS`;
			let signature = `ID=${this.imei}`
			if (this.ack == "disabled") signature = `SA=0,0;${signature}`
			let salut = `>RXART;${version};EG25S;${additional};${signature}<`;
			this.sendWithoutConfirmation(salut);
			this._sendBuffer()
		};

		// hook to channel's onMessage
		this.client.onMessage = data => {
			// continue buffer
			if (this.taip_buffer.length){
				clearTimeout(this.taip_buffer_timeout)
				data = `${this.taip_buffer}${data}`
				this.taip_buffer = ''
			}
			let buffer = data.split('>').map(i => i.split('<')).reduce((a, b) => a.concat(b))
			let messages = []
			buffer.forEach(message => {
				if (!message.trim().length) return
				if (message == this.imei) return messages.push(message)
				if (data.includes(`>${message}<`)) return messages.push(`>${message}<`)
				this.taip_buffer = `>${message}`
			})
			// console.log({taip_buffer: this.taip_buffer, messages})
			if (this.taip_buffer.length){
				clearTimeout(this.taip_buffer_timeout)
				this.taip_buffer_timeout = setTimeout(() => {
					console.log('taip:buffered', this.taip_buffer)
					this._handleData(this.taip_buffer)
					this.taip_buffer = ''
				}, 1000 * 3)
			}
			messages.map(m => this._handleData(m))
		};

		this.client.onDisconnect = () => {
			console.log("connection lost", this.name);
			this.connected = false;
		};

		this.client.keepAliveMessage = this.imei;
		this.client.connect();
		this.__ready()
	}

	/**
	 * default send command this is called from the main, process the event To the protocol
	 * @param {Object<any>} variables syrusjs_variables copy
	 * @param {Object<any>} opts event defintion parameters, min_duration, trigger rater,etc
	 * @param {Object<any>} fields fieldset from the event
	 * @param {string} label the label from the event
	 * @param {string} code the code from the event
	 */
	async send(variables, opts, fields, label, code) {
		// console.log({variables, opts, fields, label, code})
		await this.ready
		this.fields = fields; // saving a copy from the fieldset to allow taip changes the counters when taip required
		if (!opts.ack) opts.ack = this.ack
		if ("off".indexOf(opts.ack) > -1) opts.ack = "disabled"
		if ("on,enabled".indexOf(opts.ack) > -1) opts.ack = "seq"
		let taip_message = await this.JSONtoMessage(variables, code, label, fields, opts);
		this._send(taip_message);
	}

	/**
	 * adds messages to queue
	*/
	_send(message) {
		this.queue.push(message);
		if (this.queue.length == 1 && this.connected) this._sendBuffer()
	}

	// Used for pushing at event at the top of queue
	// no too much usage right now useful for debug
	sendPriority(message) {
		this.queue.unshift(message);
		if (this.queue.length == 1 && this.connected) this._sendBuffer()
	}

	// function to send command without waiting a response like QPV -> RPV
	sendWithoutConfirmation(message) {
		this.client.send(message);
	}

	// Called to dequeue an event its called after every ack or when the connection is established
	async _sendBuffer() {
		await this.queue.ready()
		if (this.queue.length == 0) return
		let message = this.queue[0]
		// discard invalid messages
		if (!message || message.includes("ID=undefined")){
			this.queue.shift()
			this._sendBuffer()
			return
		}
		// set retry
		clearTimeout(this.retry);
		this.retry = setTimeout(() => {
			console.log(`no:ack: ${++this.retries}`, message);
			if (this.client.no_disconnects || this.retries < 3) return this._sendBuffer()
			this.retries = 0
			this.client.disconnect();
			return
		}, 1000 * 30);
		// send message
		this.client.send(message, (error) => {
			if (error){
				console.error(error)
				this.client.disconnect()
				return
			}
			// keep processing queue if no ack is needed
			if (message.includes(';SA=0,')) return this._handleData(this.imei);
		});
	}

	/**
	 * Manage all the TAIP commands received from the server
	 * @param {*} data
	 */
	_handleData(data) {
		// ACK messages(imei or seq)
		if (data.trim().startsWith(this.imei) || data.includes(">SAK;")){
			clearTimeout(this.retry);
			this.retry = null;

			// no more events, invalid message
			let message = this.queue[0]
			if (this.queue.length == 0 || !message) return

			// ack checks
			// trip confirmation
			if (data.includes(">SAK;t") && !message.includes(">RXATN")) return
			let ack_required = message.includes(";SA=1,")
			if(ack_required){
				let ack_id = message.split(";SA=1,")[1].split(';')[0]
				if (!data.includes(`>SAK;1,${ack_id},`)) return this._sendBuffer()
			}
			// send photos after PS tag is acknowledged
			if (message.includes(";PS")) this.sendPhoto()
			// pop event from queue
			this.retries = 0
			this.queue.shift();
			this._sendBuffer();
			return;
		}

		// Queries
		if (data.indexOf(">QXART") != -1) {
			System.info()
				.then(info => {
					let version = info.apexVersion ? info.apexVersion.replace("apex-", "") : "00.00.00";
					version = version.split("-")[0] + ".00";
					this.sendWithoutConfirmation(this.msgbuilder(this.RPV(`>RXART;${version};EG25S;`), data));
				})
				.catch(err => {
					console.error(err);
					let version = "00.00.00";
					this.sendWithoutConfirmation(this.msgbuilder(this.RPV(`>RXART;${version};EG25S;`), data));
				});
			return;
		}
		if (data.indexOf(">QPV") != -1) {
			this.sendWithoutConfirmation(this.msgbuilder(this.RPV(this.last_data), data));
			return;
		}
		if (data.indexOf(">QVR") != -1) {
			this.sendWithoutConfirmation(this.msgbuilder(`>RVR7.0.${1}`, data));
			return;
		}
		if (data.indexOf(">QXARS") != -1) {
			this.sendWithoutConfirmation(this.msgbuilder(`RXARS;${this.IMSI};${this.carrier};${this.ICCID.replace("+QCCID: ", "")}`, data));
			return;
		}
		if (data.indexOf(">QXAEM0") != -1) {
			// using await becomes painful upstream
			Utils.OSExecute("apx-ecu list_parameters").then((ecu_parameters) => {
				let xaem_groups = [
					new Array(64).fill(0),
					new Array(64).fill(0),
					new Array(64).fill(0),
				]
				for (const ecu_key of Object.keys(ecu_parameters)){
					let ecu_item = ECU_LIST[ecu_key]
					if (!ecu_item) continue
					let { $taip_xaem_group, $taip_xaem_bit } = ecu_item
					if (!$taip_xaem_group || !$taip_xaem_bit) continue
					xaem_groups[$taip_xaem_group - 1][$taip_xaem_bit - 1] = 1
				}
				let ecu_response = xaem_groups.map(group => parseInt(group.join(''), 2)).join(',')
				this.sendWithoutConfirmation(this.msgbuilder(`RXAEM0T,60,2,${ecu_response},6.1.3-apex`, data));
			})
			return;
		}
		if (data.indexOf(">QXATT") != -1) {
			if (this.trips_events) this.sendWithoutConfirmation(this.msgbuilder("RXATT010000", data));
			else this.sendWithoutConfirmation(this.msgbuilder("RXATTU", data));
			return;
		}
		if (data.indexOf(">QSSIP") != -1) {
			let output = parseInt(data[6]);
			IOS.getInputState(`IN${output}`).then(state => {
				this.sendWithoutConfirmation(this.msgbuilder(">RSSIP" + output + state, data));
			});
			return;
		}
		if (data.indexOf(">QSSXP") != -1) {
			output = parseInt(data[6]);
			IOS.getInputState(`OUT${output}`).then(state => {
				this.sendWithoutConfirmation(this.msgbuilder(">RSSXP" + output + state, data));
			});
			return;
		}
		if (data.indexOf(">QFP") != -1) {
			this.sendWithoutConfirmation(`>RFPA;ID=${this.imei}<`);
			return;
		}
		if (data.indexOf(">QXASMS") != -1) {
			Serial.getSerialModemState().then(state => {
				let connected = state.state == "connected" ? 1 : 0
				let is_satcom = state.type == "satcom"
				let msg = `RXASMS${connected}`
				state.msg_err = state.msg_err ? state.msg_err : ''
				if (is_satcom && connected){
					msg += `${state.signal}${state.id}${state.msg_err}`
				}
				this.sendWithoutConfirmation(this.msgbuilder(msg, data));
			}).catch((error)=>{
				console.error(error)
				this.sendWithoutConfirmation(this.msgbuilder("RXASMS0", data));
			});
			return
		}
		if (data.indexOf(">QXASF00") != -1) {
			this.sendWithoutConfirmation(
				this.msgbuilder(`RXASF00${this.queue.length}+[0F]M,0B,0%`, data)
			);
			return;
		}

		// Setters
		if (data.indexOf(">SSL") != -1) {
			this.onCommand({
				isSyrusLang: true,
				command: data
					.replace(">SSL", "")
					.split("<")[0]
					.split(";ID=")[0]
					.split(";SI=")[0]
			})
				.then(response => {
					if (!response) response = "OK";
					this.sendWithoutConfirmation(
						this.msgbuilder(
							`RSL${
								typeof response == "string"
									? response.substring(0, 512).replace(/\n/g, "")
									: JSON.stringify(response).substring(0, 512).replace(/\n/g, "")
							}`,
							data
						)
					);
				})
				.catch(err => {
					let code = err.error ? err.error.code : "1";
					let errorText = JSON.stringify(err);
					if (err.errorText) {
						errorText = err.errorText;
					} else if (err.output) {
						errorText = err.output;
					}
					try {
						let newErr = JSON.parse(errorText).error;
						if (newErr) errorText = `${newErr}`;
					} catch (error) {}
					this.sendWithoutConfirmation(this.msgbuilder(
						`RTX;${code};${errorText.replace(/\n/g, "").substring(0, 512)}`, data)
					);
				});
			return;
		}
		if (data.indexOf(">SXATT01") != -1) {
			this.sendWithoutConfirmation(this.msgbuilder("RXATT01", data));
			this.trips_events = true;
			storage.set(`${this.name}_trips_events`, true);
			return;
		}
		if (data.indexOf(">SXATTU0") != -1) {
			this.sendWithoutConfirmation(this.msgbuilder("RXATTU0", data));
			this.trips_events = false;
			storage.set(`${this.name}_trips_events`, false);
			return;
		}
		if (data.indexOf(">SSSXP") != -1) {
			let output = parseInt(data[6]);
			let status = parseInt(data[7]);
			console.info(`OUT${output} changed to ${status}`);
			this.onCommand({
				command: "set out",
				io: output,
				state: status == 1 ? "on" : "off"
			});
			this.sendWithoutConfirmation(this.msgbuilder(">RSSXP" + output + status, data));
			return;
		}
		if (data.indexOf(">SSSXE") != -1) {
			// Handle this increased 2 because we used expaned outputs for additionals outputs
			let output = parseInt(data[6]);
			let status = parseInt(data[7]);
			console.info(`OUT${output + 2} changed to ${status}`);
			this.onCommand({
				command: "set out",
				io: output + 2,
				state: status == 1 ? "on" : "off"
			});
			this.sendWithoutConfirmation(this.msgbuilder(">RSSXE" + output + status, data));
			return;
		}
		if (data.indexOf(">SXACT0") != -1) {
			let evtxt = data.substring(data.indexOf(">SXACT0") + 7, data.indexOf(">SXACT0") + 9);
			this.sendWithoutConfirmation(this.msgbuilder(">RXACT0" + evtxt, data));
			this.onCommand({
				command: "send event",
				code: evtxt,
				label: "trckpnt"
			});
			return;
		}
		if (data.indexOf(">SXAVOV") != -1) {
			let newval = pad(Number(data.substring(7, data.indexOf(";", 7))), 9);
			this.adjustCounter("VO", newval);
			let msg = this.msgbuilder("RXAVOV" + newval, data);
			this.sendWithoutConfirmation(msg);
			return;
		}
		if (data.indexOf(">SXACLV") != -1) {
			let newval = pad(Number(data.substring(7, data.indexOf(";", 7))), 9);
			this.adjustCounter("CL", newval);
			let msg = this.msgbuilder("RXACLV" + newval, data);
			this.sendWithoutConfirmation(msg);
			return;
		}
		if (data.indexOf(">SXACEV") != -1) {
			let newval = pad(Number(data.substring(7, data.indexOf(";", 7))), 9);
			this.adjustCounter("CE", newval);
			let msg = this.msgbuilder("RXACEV" + newval, data);
			this.sendWithoutConfirmation(msg);
			return;
		}
		if (data.indexOf(">SXACSV") != -1) {
			let newval = pad(Number(data.substring(7, data.indexOf(";", 7))), 9);
			this.adjustCounter("CS", newval);
			let msg = this.msgbuilder("RXACSV" + newval, data);
			this.sendWithoutConfirmation(msg);
			return;
		}
		if (data.indexOf(">SXACRV") != -1) {
			let newval = pad(Number(data.substring(7, data.indexOf(";", 7))), 9);
			this.adjustCounter("CR", newval);
			let msg = this.msgbuilder("RXACRV" + newval, data);
			this.sendWithoutConfirmation(msg);
			return;
		}
		if (data.indexOf(">SXASC") != -1) {
			if (data.indexOf(">SXASCM0") != -1) {
				this.ftp = {}
				storage.remove(`${this.name}_ftp_config`);
				this.sendWithoutConfirmation(this.msgbuilder("RXASCM0", data));
			}
			if (data.indexOf(">SXASCM202") != -1){
				// Activate camera Action: Prepare Yourself
				this.sendWithoutConfirmation(this.msgbuilder(">RXASCM202", data));
			}
			if (data.indexOf(">SXASCP99") != -1) {
				Utils.OSExecute('apx-serial fatigue_sensor capture_photo')
				.then(response => {
					this.sendWithoutConfirmation(this.msgbuilder(">RXASCP99", data));
				})
				.catch(err => {
					let code = err.error ? err.error.code : "1";
					let errorText = JSON.stringify(err);
					this.sendWithoutConfirmation(this.msgbuilder(
						`RXAEC::${command};${code};${errorText.replace(/\n/g, "").substring(0, 512)}`, data)
					);
				})
			}
			return;
		}
		if (data.indexOf(">SSSU051") != -1) {
			Utils.OSExecute("apx-serial fatigue_sensor capture_photo")
			this.sendWithoutConfirmation(this.msgbuilder(">RSSU051", data));
			return;
		}
		if (data.indexOf(">SSSU061") != -1) {
			this.sendWithoutConfirmation(this.msgbuilder(">RSSU061", data));
			this.onCommand({
				command: "send event",
				code: "19",
				label: "position"
			});
			return;
		}
		// MDT
		if (data.indexOf(">STX") != -1) {
			let mdt = data.split(">STX")[1].split(";")[0]
			Utils.OSExecute(`apx-serial mdt send '${mdt}'`)
			.then(response =>{
				this.sendWithoutConfirmation(
					this.msgbuilder(
						`RTX${mdt}`,
						data
					)
				);
			})
			.catch(error =>{
				this.sendWithoutConfirmation(
					this.msgbuilder(
						`RER00;${data
							.split(";")[0]
							.replace(">", "")
							.replace("<", "")}`,
						data
					)
				)
			})
			return;
		}
		// FTP
		if (data.indexOf(">SXAFSA1") != -1) {
			//>SXAFSA1dev.pegasusgateway.com:7021;ID=450000210622036;SI=5-28bfc<
			let split = data
				.split(">SXAFSA1")[1]
				.split(";")[0]
				.split(":");
			this.ftp.server = split[0];
			this.ftp.port = split[1];
			this.sendWithoutConfirmation(this.msgbuilder("RXAFSA1" + this.ftp.server + ":" + this.ftp.port, data));
			storage.hset(`${this.name}_ftp_config`, 'server', split[0]);
			storage.hset(`${this.name}_ftp_config`, 'port', split[1]);
			return
		}
		if (data.indexOf(">SXAFSL") != -1) {
			this.ftp.username = data.split(">SXAFSL")[1].split(";")[0];
			this.sendWithoutConfirmation(this.msgbuilder("RXAFSL" + this.ftp.username, data));
			storage.hset(`${this.name}_ftp_config`, 'username', this.ftp.username);
			return
		}
		if (data.indexOf(">SXAFSP") != -1) {
			this.ftp.password = data.split(">SXAFSP")[1].split(";")[0];
			this.sendWithoutConfirmation(this.msgbuilder("RXAFSP" + this.ftp.password, data));
			storage.hset(`${this.name}_ftp_config`, 'password', this.ftp.password);
			return
		}
		// END FTP
		if (data.indexOf(">SXABR9600") != -1) {
			this.sendWithoutConfirmation(this.msgbuilder("RXABR9600", data));
			return;
		}
		if (data.indexOf(">SXAIBC") != -1) {
			// this.events.publish("unlink", {});
			this.sendWithoutConfirmation(this.msgbuilder("RXAIBC", data));
			return;
		}
		if (data.indexOf(">SRT;SFBUFF") != -1) {
			this.queue.clear();
			this.sendWithoutConfirmation(this.msgbuilder("RRT;SFBUFF", data));
			return;
		}
		if (data.indexOf(">SXAEC::") != -1) {
			let command = data
				.replace(">SXAEC::", "")
				.split("<")[0]
				.split(";ID=")[0]
				.split(";SI=")[0];
			console.log("executed remote command", command);
			Utils.OSExecute(command)
				.then(response => {
					if (!response) response = "OK";
					this.sendWithoutConfirmation(
						this.msgbuilder(
							`RXAEC::${command.replace(/\n/g, "")};0;${
								typeof response == "string"
									? response.substring(0, 512).replace(/\n/g, "")
									: JSON.stringify(response).substring(0, 512).replace(/\n/g, "")
							}`,
							data
						)
					);
				})
				.catch(err => {
					let code = err.error ? err.error.code : "1";
					let errorText = JSON.stringify(err);
					if (err.errorText) {
						errorText = err.errorText;
					} else if (err.output) {
						errorText = err.output;
					}
					try {
						let newErr = JSON.parse(errorText).error;
						if (newErr) errorText = `${newErr}`;
					} catch (error) {}
					this.sendWithoutConfirmation(this.msgbuilder(
						`RXAEC::${command};${code};${errorText.replace(/\n/g, "").substring(0, 512)}`, data)
					);
				});
			return;
		}

		/**
		 * Garmin Mode Implementation
		 */
		if (data.indexOf(">QXAGM") != -1 || data.indexOf(">SXAGM") != -1) {
			this._handlerGarmin(data);
			return;
		}

		// Catch all unhadled command
		this.sendWithoutConfirmation(
			this.msgbuilder(
				`RER00;${data
					.split(";")[0]
					.replace(">", "")
					.replace("<", "")}`,
				data
			)
		);
	}

	/**
	 * this function is intended for handling data received for Garmin plugin from syrus3, not used right now but keep it here for reference
	 * @param {string} data the taip data
	 */
	_handlerGarmin(data) {
		// the gateway ask for the status of garmin
		if (data.indexOf(">QXAGMI") != -1) {
			this.sendWithoutConfirmation(this.msgbuilder(`>RXAGMI${this.imei.substring(5)}00001${pad("00570", 5)}0000`, data));
			return;
		}

		// The Gateway ask if the garmin device is connected
		if (data.indexOf(">SXAGME0") != -1) {
			this.sendWithoutConfirmation(this.msgbuilder(">RXAGME0", data));
			return;
		}

		/** The gateway send a text message */
		if (data.indexOf(">SXAGMTS") != -1) {
			this.sendWithoutConfirmation(this.msgbuilder(data.split(";")[0].replace(">SXAGM", ">RXAGM"), data));
			// let id = parseInt(data.split(";")[0].substring(8, 13));
			// let timestamp = data.split(";")[0].substring(13, 24);
			// let msg = data
			// 	.split(";")[0]
			// 	.substring(24)
			// 	.autoLink({ target: "_blank", rel: "nofollow" });
		}

		// Cath all unhadled command from Garmin mode
		if (data.indexOf(">SXAGM") != -1) {
			this.sendWithoutConfirmation(this.msgbuilder(data.split(";")[0].replace(">SXAGM", ">RXAGM"), data));
			return;
		}
	}

	/**
	 * Converts Json Data from the events in a TAIP event text
	 *
	 * @param {*} data Json Data received from the syrusjs variables and fieldsets
	 * @returns TAIP text message
	 *
	 * @memberOf Parser
	 */
	async JSONtoMessage(data, code = "01", label = "trckpnt", fields, opts) {
		this.last_data = data;
		let weeks = -dayjs.unix(315964800).diff(dayjs.utc(), "weeks");
		let day = dayjs.utc().day();
		let seconds = pad(dayjs.utc().diff(dayjs.utc().startOf("day"), "seconds"), 5);
		let lat = getLat(data["$gnss.latitude"] || 0);
		let long = getLong(data["$gnss.longitude"] || 0);
		let speed = data["$gnss.speed"] >= 0 ? pad(Math.round(Math.abs(data["$gnss.speed"] * 2.23694)), 3) : "000";
		let heading = data["$gnss.bearing"] >= 0 ? pad(Math.round(Math.abs(data["$gnss.bearing"])), 3) : "000";
		let k = data["$gnss.fix"] || "0";
		let l = ""
		if (!data["$gnss.timestamp"]) l = "0";
		else l = dayjs(data["$gnss.timestamp"]).diff(dayjs(), "seconds") < 10 ? "2" : "1";
		let altitude = data["$gnss.altitude"] ? ";AL=" + Math.floor(data["$gnss.altitude"]) : "";
		let dop = "";
		if (data["$gnss.hdop"]) {
			dop = `;DOP=${parseInt(data["$gnss.pdop"] * 100)},${parseInt(data["$gnss.hdop"] * 100)},${parseInt(data["$gnss.vdop"] * 100)}`;
		}
		let bl = "";
		if (data["$battery.mv"] && Math.floor(data["$battery.mv"]) > -1) bl = ";BL=" + Math.floor(data["$battery.mv"]);
		let evLabel = ";EVLBL=" + label;
		let ix = "";
		let adaw = "";
		if(fields["$io"]){
			ix = getIO(data);
			adaw = getAnalogs(data);
		}
		let sc = await getSC(fields, data);
		let satcom = await getSatcom(fields, data);
		let imeipart = ";ID=" + this.imei;
		let ackpart = ""
		if (opts.ack == "disabled"){
			ackpart = `;SA=0,${this.ack_counter}`
		}
		if (opts.ack == "seq"){
			ackpart = `;SA=1,${this.ack_counter}`
			await storage.set(`${this.name}_ack`, ++this.ack_counter);
		}
		let aidpart = "";
		let sv = "";
		if (data["$gnss.satsActive"]) sv = `;SV=${data["$gnss.satsActive"]}`;
		code = pad(code, 2);
		let ecu = "";
		if(fields["$ecu"]){
			ecu = ECUToTAIP(data);
		}
		let counters = getCounters(fields, data);
		let ib_is = getIB_IS(fields, data)
		let ea = getEA(fields, data)

		let taip = ""

		// photo
		let photo = await this.getPS(fields, data, opts)
		// MDT
		if (opts.trigger.includes('@mdt.message')){
			let mdt = data['$mdt.message']
			taip += `>RTX${mdt};EV`
		}
		if (!taip.length) taip += `>REV`
		taip += `${code}${weeks}${day}${seconds}${lat}${long}${speed}${heading}${k}${l}${altitude}`
		taip += `${evLabel}${dop}${bl}${ix}${adaw}${sv}${ecu}${counters}${ib_is}${ea}${aidpart}${sc}`
		taip += `${satcom}`
		taip += `${photo}`
		taip += `${ackpart}${imeipart}<`;
		return taip;
	}

	/**
	 * RPV taip constructor
	 * @param {string} data the variables from syrusjs
	 */
	RPV(data = {}) {
		let seconds = pad(dayjs.utc().diff(dayjs.utc().startOf("day"), "seconds"), 5);
		let lat = getLat(data["$gnss.latitude"] || 0);
		let long = getLong(data["$gnss.longitude"] || 0);
		let speed = data["$gnss.speed"] >= 0 ? pad(Math.round(Math.abs(data["$gnss.speed"] * 2.23694)), 3) : "000";
		let heading = data["$gnss.bearing"] >= 0 ? pad(Math.round(Math.abs(data["$gnss.bearing"])), 3) : "000";
		let k = data["$gnss.fix"] || 0;
		let l
		if (!data["$gnss.timestamp"]) l = 0;
		else l = dayjs(data["$gnss.timestamp"]).diff(dayjs(), "seconds") < 10 ? "2" : "1";
		return `>RPV${seconds}${lat}${long}${speed}${heading}${k}${l}`;
	}

	/**
	 * Converts TAIP Data to JSON syrus.js procesable as order
	 *
	 * @param {*} TAIP Data received from the listener
	 * @returns JSON message order
	 *
	 */
	MessagetoJSON(taip) {}

	/**
	 * helper to construct response to gateway, that return the TAIP command with ID, SI
	 *
	 * @param {string} message string that will be send to gateway without ID or SI parts
	 * @param {string} original the original message received from gateway
	 * @returns {string} the TAIP text command to send to gateway
	 *
	 */
	msgbuilder(message, original) {
		if (!original) original = ''
		let text = "";
		if (message[0] != ">") {
			text = ">";
		}
		text += message;
		if (original.indexOf(";SI=") != -1) {
			text += original.substring(original.indexOf(";SI="), original.length - 1);
		}
		if (text.indexOf(";ID=") == -1) {
			text += ";ID=" + this.imei;
		}
		if (text.indexOf("<") == -1) {
			text += "<";
		}
		return text;
	}

	/**
	 * allow to adjust counters from the fieldset defined
	 * @param {String} key the value to adjust
	 * @param {Number} newValue the new value to set
	 */
	async adjustCounter(key, newValue) {
		const DEFAULTS_COUNTERS = {
			VO: "$counters.globals.odometer",
			CL: "$counters.globals.idle_time",
			CE: "$counters.globals.ignition_time",
			CS: "$counters.globals.over_speed",
			CR: "$counters.globals.over_rpm"
		};
		let definition = null
		if (this.fields) definition = this.fields[key];
		if (!definition) definition = DEFAULTS_COUNTERS[key];
		let [signal, namespace, value] = definition.split(".");
		if (signal !== "$counters") console.error(`${key} is not assigned to an apx counter`);
		let promise = Utils.OSExecute(`apx-counter set ${namespace} ${value.toUpperCase()} ${Math.round(Number(newValue))}`);
		promise.catch(console.error);
		return promise;
	}

	// Photo
	async getPS(fields, data, opts) {
		// console.log({fields, data, opts});
		// console.log('photo_state', this.photo_state)
		// return ""
		let tag = ""
		const { video, photo } = opts
		if (video) {
			let media_type = 100
			let media_id = (parseInt(Date.now() / 1e3) % 1e8) * 10
			let sources = video.split(',')
			for (const source of sources) {
				// find video element
				const video = {}
				const { system_epoch, camera, file } = video
				if (system_epoch <= this.photo_state[source]) continue
				const media_dest = `${this.imei}/images/${media_id}.mp4`
				this.photo_state[file.path] = system_epoch
				storage.hset(`${this.name}_photo_state`, file.path, system_epoch)
				this.queuePhoto(media_id, media_type, file.path, media_dest)
				tag += `;PS${media_type}=${media_id}`
				media_type++
				media_id++
			}
		}
		while (photo) {
			let source = photo
			let key
			if (source.startsWith('fatigue_sensor')) {
				key = "$fatigue_sensor.latest_photo"
				if (source.includes('.')) {
					key = `$fatigue_sensor.photos.${source.split('.')[1]}`
				}
			} else break

			let photo_src = data[key]
			if (!photo_src) break
			let photo_epoch = photo_src.split('-')[0]
			let last = this.photo_state[key]
			if (photo_epoch <= last) break
			if (source.startsWith('fatigue_sensor')) {
				photo_src = `/data/users/syrus4g/fatigue_sensor/${photo_src}`
			}
			let photo_type = "00"
			let photo_id = parseInt(Date.now() / 1e3) % 1e7
			let photo_dest = `${this.imei}/images/${photo_id}.jpeg`
			this.photo_state[key] = photo_epoch
			storage.hset(`${this.name}_photo_state`, key, photo_epoch)
			// schedule upload, add PS tag
			this.queuePhoto(photo_id, photo_type, photo_src, photo_dest)
			tag += `;PS${photo_type}=${photo_id}`
			break
		}
		return tag
	}

	/**
	 * Queue photo for upload to FTP
	 * @param {*} photo_id
	 * @param {*} filename
	 * @memberof TAIP
	 */
	async queuePhoto(photo_id, photo_type, photo_src, photo_dest){
		// push photo to queue
		let queue = `${this.name}_photo_queue`
		await storage.push(queue, `${photo_id}::${photo_type}::${photo_src}::${photo_dest}`)
	}
	/**
	 * uploads photo, sends respective SXACN commands
	 * @memberof TAIP
	 */
	async sendPhoto(){
		let queue = `${this.name}_photo_queue`
		let photos = await storage.slice(queue, -1, -1)
		if (!photos.length) return
		let photo = photos[0]
		let [photo_id, photo_type, src, dest] = photo.split('::')
		try {
			this.sendWithoutConfirmation(this.msgbuilder(`RXASCN3,${photo_id},${photo_type}`))
			let cmd = ""
			cmd += `lftp -u ${this.ftp.username},${this.ftp.password} -p ${this.ftp.port}`
			cmd += ` -e 'put ${src} -o ${dest}; bye'`
			cmd += ` ftp://${this.ftp.server}`
			console.log(cmd)
			await Utils.OSExecute(cmd)
			this.sendWithoutConfirmation(this.msgbuilder(`RXASCN4,${photo_id},${photo_type}`))
		} catch(error) {
			console.error(error)
		}
		await storage.pop(queue)
		this.sendPhoto()
		return
	}
}
